/* Jean Philippe EIMER - 2013 *
 *      phil@jpmr.org         */

#include <math.h>

#include <iostream>
#include <iomanip>
using namespace std;

#include "include.h"


void	Body::Init(Ephemeris *pEphe, Orbit::PLANET id)
{
	// Geocentric Coordinates for The Sun and The Moon
	// Planets' Orbital Elements
	static const wxString	Vnom[NB_ASTR] = { \
			MER_NAME, VEN_NAME, \
			MAR_NAME, JUP_NAME, SAT_NAME, \
			URA_NAME, NEP_NAME, /*PLU_NAME,*/ \
			EAR_NAME, SUN_NAME, MOO_NAME};


	pE = pEphe;
	bodyid = id;
	name = Vnom[id];

	OrbitElem.Init(this);

	Rising.tz = pE->Observ.Inst.tz;
	Meridian.tz = pE->Observ.Inst.tz;
	Setting.tz = pE->Observ.Inst.tz;

	if( id == Orbit::EAR )
		Coord[Ephemeris::HELIO] = OrbitElem2EcliptHelio(OrbitElem);

	Parallax();
}


Vector Body::EcliptGeo(void)
{
	Vector	Vout;


	OrbitElem.Elements(pE->Observ.Inst.t);
	OrbitElem.Perturbations(pE->Observ.Inst.t);
	OrbitElem.Anomalies();

	Vout=OrbitElem2EcliptHelio(OrbitElem);

	if( bodyid != Orbit::SUN )
	{
		Coord[Ephemeris::HELIO]=Vout;
		Vout=EcliptHelio2EcliptGeo(Vout,pE->Sun.Coord[Ephemeris::GEO]);
	}

	Coord[Ephemeris::GEO]=Vout;

	return(Vout);
}


Vector	Body::EcliptGeo2Horizon(double t)
{
	Coord[Ephemeris::GEOAPP]=EcliptGeo2EcliptGeoApp(Coord[Ephemeris::GEO],pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_TETA],t);
	Coord[Ephemeris::EQUA]=EcliptGeoApp2Equa(Coord[Ephemeris::GEOAPP]);
	Coord[Ephemeris::HOUR]=Equa2Hour(Coord[Ephemeris::EQUA]);
	Coord[Ephemeris::HORIZON]=Hour2Horizon(Coord[Ephemeris::HOUR]);

	return(Coord[Ephemeris::HORIZON]);
}


Vector Body::OrbitElem2EcliptHelio(const Orbit &Orb) const
{
	Vector	Vout;


	Vout.Spher[Vector::COORD_R]=Orb.r;
	Vout.Spher[Vector::COORD_TETA]=Orb.omega+Orb.v;
	Vout.Spher[Vector::COORD_PHI]=0;

	Vout.Rectangular();

	Vout.RotationX(-Orb.i);

	Vout.Spher[Vector::COORD_TETA]+=Orb.Omega;

	Vout.Spher[Vector::COORD_R]+=Orb.rper;		// Corrections
	Vout.Spher[Vector::COORD_TETA]+=Orb.lper;
	Vout.Spher[Vector::COORD_PHI]+=Orb.bper;

	Vout.Rectangular();

	return(Vout);
}


Vector Body::EcliptHelio2EcliptGeo(const Vector &Vin,const Vector &GeoSol) const
{
	Vector V1=Vin,V2=GeoSol,Vout;

	V1.Spher[Vector::COORD_TETA]-=GeoSol.Spher[Vector::COORD_TETA];		// l-lamdaS
	V1.Rectangular();

	V2.Spher[Vector::COORD_TETA]=0;
	V2.Spher[Vector::COORD_PHI]=0;
	V2.Rectangular();

	Vout=V1+V2;

	Vout.Spher[Vector::COORD_TETA]+=GeoSol.Spher[Vector::COORD_TETA];	// l+lamdaS = lamda
	Vout.Rectangular();

	return(Vout);	// d, lamda and beta
}


Vector Body::EcliptGeo2EcliptHelio(const Vector &Vin,const Vector &GeoSol) const
{
	Vector V1=Vin,V2=GeoSol,Vout;

	V1.Spher[Vector::COORD_TETA]-=GeoSol.Spher[Vector::COORD_TETA];		// lamda-lamdaS
	V1.Rectangular();


	V2.Spher[Vector::COORD_TETA]=0;
	V2.Spher[Vector::COORD_PHI]=0;
	V2.Rectangular();

	Vout=V1-V2;

	Vout.Spher[Vector::COORD_TETA]+=GeoSol.Spher[Vector::COORD_TETA];	// lamda+lamdaS = l
	Vout.Rectangular();

	return(Vout);	// r, l and b
}


Vector Body::EcliptGeo2EcliptGeoApp(const Vector &Vin,double lamdaS,double t) const
{
	Orbit	Orb=OrbitElem;
	Vector	Vout;
	double	lsl,tl,dL;


	if(bodyid < Orbit::SUN)
	{
		// Light Travel
		tl=LightTime(Vin.Spher[Vector::COORD_R]);
		dL=tl*Orb.deltaL;
		Orb.L-=dL;				// Mean Longitude Modification
		Orb.MRsp-=DEG2RAD(dL);	// Mean Anomaly Modification

		if( bodyid < Orbit::JUP )
		{
			Orb.L-=Orb.Lper;
			Orb.Perturbations(t-tl);
		}
		else
			Orb.Lper=Orb.vkper=Orb.eper=Orb.aper=0;

		Orb.Anomalies();

		Vout=EcliptHelio2EcliptGeo(OrbitElem2EcliptHelio(Orb),pE->Sun.Coord[Ephemeris::GEO]);
	}
	else
		Vout=Vin;

	Vout.Spher[Vector::COORD_TETA]+=pE->Nutat.lamda;

	if( bodyid != Orbit::MOO )
	{
		lsl=DEG2RAD(lamdaS-Vin.Spher[Vector::COORD_TETA]);
		Vout.Spher[Vector::COORD_TETA]-=COEF_APP*cos(lsl)/cos(DEG2RAD(Vin.Spher[Vector::COORD_PHI]));
		Vout.Spher[Vector::COORD_PHI]-=COEF_APP*sin(lsl)*sin(DEG2RAD(Vin.Spher[Vector::COORD_PHI]));
	}

	Vout.Rectangular();

	return(Vout);	// lamda App and beta App
}


Vector Body::EcliptGeoApp2EcliptGeo(const Vector &Vin,double lamdaS) const
{
	Vector	Vout;
	double	lsl=DEG2RAD(lamdaS-Vin.Spher[Vector::COORD_TETA]);


	Vout=Vin;
	Vout.Spher[Vector::COORD_TETA]-=pE->Nutat.lamda;

	if( bodyid != Orbit::MOO )
	{
		Vout.Spher[Vector::COORD_TETA]+=COEF_APP*cos(lsl)/cos(Vin.Spher[Vector::COORD_PHI]);
		Vout.Spher[Vector::COORD_PHI]+=COEF_APP*sin(lsl)*sin(Vin.Spher[Vector::COORD_PHI]);
	}

	Vout.Rectangular();

	return(Vout);	// lamda and beta
}


Vector Body::EcliptGeoApp2Equa(const Vector &Vin) const
{
	Vector Vout;


	Vout=Vin;
	Vout.RotationX(-(pE->epsilon+pE->Nutat.epsilon));

	return(Vout);	// alpha and delta
}


Vector Body::Equa2EcliptGeoApp(const Vector &Vin) const
{
	Vector	Vout;


	Vout=Vin;
	Vout.RotationX(pE->epsilon+pE->Nutat.epsilon);

	return(Vout);	// lamda and beta
}


Vector Body::Equa2Hour(const Vector &Vin) const
{
	Vector	Vout;


	Vout=Vin;

	Vout.Spher[Vector::COORD_TETA]=pE->Observ.Inst.ts*KH2D-Vin.Spher[Vector::COORD_TETA];	// TS-alpha
	Vout.Rectangular();

	return(Vout);	// H and delta
}


Vector Body::Hour2Equa(const Vector &Vin) const
{
	Vector	Vout;


	Vout=Vin;

	Vout.Spher[Vector::COORD_TETA]=pE->Observ.Inst.ts*KH2D-Vin.Spher[Vector::COORD_TETA];	// TS-H
	Vout.Rectangular();

	return(Vout);	// alpha and delta
}


Vector Body::Hour2Horizon(const Vector &Vin) const
{
	Vector	Vout;


	Vout=Vin;
	Vout.RotationY(pE->Observ.param.location_lat-90.0);
	Vout.RotationZ(180.0);	// Azimuth has South as reference

	return(Vout);	// A and h, azimuth and height
}


Vector Body::Horizon2Hour(const Vector &Vin) const
{
	Vector	Vout;


	Vout=Vin;
	Vout.RotationZ(180.0);	// Azimuth has South as reference
	Vout.RotationY(90.0-pE->Observ.param.location_lat);

	return(Vout);			// H et delta
}


// Computes Light Travel Time
double	Body::LightTime(double dist) const
{
	return(dist*COEF_LIGHT);
}


// Computes Rising and Setting Times
void	Body::RisingSetting(void)
{
	Vector	Eq12;
	double	h0,tu,tu0,trig[3];


	h0=parallax-diameter/7200.0-pE->Observ.param.refraction-pE->Observ.eta1;

	trig[2]=DEG2RAD(pE->Observ.param.location_lat);
	trig[1]=sin(trig[2]);
	trig[2]=cos(trig[2]);

	// Rising
	tu0=12.0;	// Start at 12 h UT
	Eq12=tu2Eq(tu0);
	trig[0]=sin(DEG2RAD(h0+pE->Observ.eta2L));
	tu=Eq2tu(Eq12,trig,RISING);

	while(TUcond(&tu,tu0,PRECISION_LC))
	{
		tu0=tu;
		tu=Eq2tu(tu2Eq(tu),trig,RISING);
	}

	Rising.tu2t( pE->Observ.Inst, pE->Observ.dst, tu );

	// Setting
	tu0=12.0;	// Start at 12 h UT
	trig[0]=sin(DEG2RAD(h0+pE->Observ.eta2C));
	tu=Eq2tu(Eq12,trig,SETTING);

	while(TUcond(&tu,tu0,PRECISION_LC))
	{
		tu0=tu;
		tu=Eq2tu(tu2Eq(tu),trig,SETTING);
	}

	Setting.tu2t( pE->Observ.Inst, pE->Observ.dst, tu );

	// Meridian Crossing Time
	tu0=12.0;	// Start at 12 h UT
	tu=Eq2tu(Eq12,trig,MERIDIAN);

	while(TUcond(&tu,tu0,PRECISION_M))
	{
		tu0=tu;
		tu=Eq2tu(tu2Eq(tu),trig,MERIDIAN);
	}

	Meridian.tu2t( pE->Observ.Inst, pE->Observ.dst, tu );
}


// Computes Universal Time from height and Equatorial Coordinates
double	Body::Eq2tu(const Vector &Eq, double trig[], RSM_T lcm)
{
	double	H,tu=0;


	if(lcm==MERIDIAN)
		H=0;
	else
	{
		H=DEG2RAD(Eq.Spher[Vector::COORD_PHI]);
		H=(trig[0]-trig[1]*sin(H))/trig[2]/cos(H);

		if(H>=-1.0 && H<=1.0)
			H=RAD2DEG(acos(H));
		else
			tu=TUSING;

		if(lcm==RISING)
			H=-H;
	}

	if(tu!=TUSING)
	{
		tu=(H+pE->Observ.param.location_long+Eq.Spher[Vector::COORD_TETA])/KH2D-pE->Observ.Inst.tsg0;
		tu=K24(tu);
		tu=tu/SOL2SID;
	}

	return(tu);
}


// Computes Equatorial Coordinates from Universal Time
Vector	Body::tu2Eq(double tu)
{
	Orbit		Orb=OrbitElem;
	MoonOrbit	MoonOrb;
	Vector	Geo;
	Instant	I;


	I.tu2t( pE->Observ.Inst, pE->Observ.dst, tu );

	if( bodyid == Orbit::MOO )
	{
		MoonOrb=pE->Moon.OrbitElem;
		MoonOrb.Elements(I);
		Geo=pE->Moon.OrbitElem2EcliptGeo(MoonOrb,I.t);
	}
	else
	{
		Orb.Elements(I.t);
		Orb.Perturbations(I.t);
		Orb.Anomalies();

		Geo=OrbitElem2EcliptHelio(Orb);

		if( bodyid != Orbit::SUN )
		{
			Orb=pE->Sun.OrbitElem;
			Orb.Elements(I.t);
			Orb.Perturbations(I.t);
			Orb.Anomalies();
			Geo=EcliptHelio2EcliptGeo(Geo,OrbitElem2EcliptHelio(Orb));
		}
	}


	return(EcliptGeoApp2Equa(Geo));
}


bool	Body::TUcond(double *ptu, double tu0, double prec)
{
	bool	cond;
	double	abs;


	if(*ptu == TUSING)
		cond = false;
	else
	{
		abs=fabs(*ptu-tu0);

		if(abs>12.1)
		{
			*ptu=TUSING;
			cond = false;
		}
		else
			cond = abs>prec;
	}

	return(cond);
}


// Computes Phase
double	Body::Phase(void)
{
	phase=acos((Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]*Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]+ \
				Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]*Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]- \
				pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_R]*pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_R]) \
                /2/Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]/Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]);

	phase100=cos(phase/2.0);
	phase=RAD2DEG(phase);

	phase100*=100.0*phase100;

	return(phase);
}


// Computes Elongation
double	Body::Elongation(void)
{
	elongation=acos((pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_R]*pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_R]+ \
                Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]*Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]- \
				Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]*Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]) \
                /2/pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_R]/Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]);

	elongation=RAD2DEG(elongation);

	return(elongation);
}


// Computes Apparent Diameter
double	Body::Diameter(void)
{
	static const double	Vdia[NB_ASTR]={MER_DIA, VEN_DIA, \
								MAR_DIA, JUP_DIA, SAT_DIA, \
								URA_DIA, NEP_DIA, /*PLU_DIA,*/ 0, SUN_DIA, MOO_DIA};


	if( bodyid == Orbit::MOO )
	{
		diameter=2*asin(Vdia[Orbit::MOO]*sin(DEG2RAD(parallax)));
		diameter=RAD2DEG(diameter)*3600.0;	// in arc second
	}
	else
		diameter=Vdia[bodyid]/Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R];	// in arc second

	return(diameter);
}


// Computes Parallax
double	Body::Parallax(void)
{
	parallax=0;

	return(parallax);
}


// Computes Magnitude
double	Body::Magnitude(void)
{
	static const VectData	Vmag[NB_ASTR]={MER_VMAG,VEN_VMAG, \
				MAR_VMAG,JUP_VMAG,SAT_VMAG, \
				URA_VMAG,NEP_VMAG,/*PLU_VMAG,*/{{0,0,0,0}},SUN_VMAG,MOO_VMAG};
	VectData	Vphi;

	Vphi.d[0]=1.0;
	Vphi.d[1]=fabs(phase)/100.0;
	Vphi.d[2]=Vphi.d[1]*Vphi.d[1];
	Vphi.d[3]=Vphi.d[2]*Vphi.d[1];

	magnitude=Vphi.Dot(Vmag[bodyid]);

	if( bodyid != Orbit::SUN )
		magnitude+=5.0*log10(Coord[Ephemeris::GEOAPP].Spher[Vector::COORD_R]*Coord[Ephemeris::HELIO].Spher[Vector::COORD_R]);

	return(magnitude);
}


///////////////////////////////////////////////////////////////////////////////
void	MoonBody::Init(Ephemeris *pEphe)
{
	pE = pEphe;
	bodyid = Orbit::MOO;
	name = MOO_NAME;

	Rising.tz = pE->Observ.Inst.tz;
	Meridian.tz = pE->Observ.Inst.tz;
	Setting.tz = pE->Observ.Inst.tz;

	OrbitElem.Init(this);
}


Vector	MoonBody::EcliptGeo(void)
{
	OrbitElem.Elements(pE->Observ.Inst);
	Coord[Ephemeris::GEO]=OrbitElem2EcliptGeo(OrbitElem,pE->Observ.Inst.t,&parallax);
	Coord[Ephemeris::HELIO]=EcliptGeo2EcliptHelio(Coord[Ephemeris::GEO],pE->Sun.Coord[Ephemeris::GEO]);

	return(Coord[Ephemeris::GEO]);
}


Vector MoonBody::OrbitElem2EcliptGeo(const MoonOrbit &Orb,double t,double *pparalx)
{
	static const VectData	Vlamda[50]=MOO_L_VV,Vbeta[45]=MOO_B_VV;
	static const VectData	Vpara[31]=MOO_P_VV;
	static const double	Clamda[50]=MOO_L_VC,Cbeta[45]=MOO_B_VC;
	static const double	Cpara[31]=MOO_P_VC;
	Vector	Vout;
	VectData	V,S;
	unsigned short int	i;
	double	paralx;


	S.d[0]=sin(DEG2RAD(Orb.Va.DotT(t)));
	S.d[1]=sin(Orb.OmegaR);
	S.d[2]=MOO_PER_CB*sin(DEG2RAD(Orb.Vb.DotT(t)));
	S.d[3]=sin(Orb.CCR);
	S.d[4]=0.0;

	V.d[0]=DEG2RAD(Orb.M+Orb.VMper.Dot(S));
	V.d[1]=DEG2RAD(Orb.F+Orb.VFper.Dot(S));
	V.d[2]=DEG2RAD(Orb.D+Orb.VDper.Dot(S));
	V.d[3]=pE->Sun.OrbitElem.MRsp+DEG2RAD(MOO_PER_CMS*S.d[0]);
	V.d[4]=0.0;

	//************************************
	// Computes Geocentric Longitude
	Vout.Spher[Vector::COORD_TETA]=0;
	for(i=45;i<50;i++)
		Vout.Spher[Vector::COORD_TETA]+=Clamda[i]*sin(V.Dot(Vlamda[i]));

	Vout.Spher[Vector::COORD_TETA]*=Orb.EE;

	for(i=27;i<45;i++)
		Vout.Spher[Vector::COORD_TETA]+=Clamda[i]*sin(V.Dot(Vlamda[i]));

	Vout.Spher[Vector::COORD_TETA]*=Orb.EE;

	for(i=0;i<27;i++)
		Vout.Spher[Vector::COORD_TETA]+=Clamda[i]*sin(V.Dot(Vlamda[i]));

	Vout.Spher[Vector::COORD_TETA]+=Orb.L+Orb.VLper.Dot(S);

	//***********************************
	// Computes Geocentric Latitude
	Vout.Spher[Vector::COORD_PHI]=Orb.EE*Cbeta[44]*sin(V.Dot(Vbeta[44]));

	for(i=29;i<44;i++)
		Vout.Spher[Vector::COORD_PHI]+=Cbeta[i]*sin(V.Dot(Vbeta[i]));

	Vout.Spher[Vector::COORD_PHI]*=Orb.EE;

	for(i=0;i<29;i++)
		Vout.Spher[Vector::COORD_PHI]+=Cbeta[i]*sin(V.Dot(Vbeta[i]));

	Vout.Spher[Vector::COORD_PHI]*=(1.0-MOO_PER_CW1*cos(Orb.OmegaR)-MOO_PER_CW2*cos(Orb.CCR));

	//**********************
	// Computes Parallax
	paralx=Orb.EE*Cpara[30]*cos(V.Dot(Vpara[30]));

	for(i=18;i<30;i++)
		paralx+=Cpara[i]*cos(V.Dot(Vpara[i]));

	paralx*=Orb.EE;
	for(i=0;i<18;i++)
		paralx+=Cpara[i]*cos(V.Dot(Vpara[i]));

	Vout.Spher[Vector::COORD_R]=EAR_RT/sin(DEG2RAD(paralx))/KUA2KM;
	Vout.Rectangular();

	if(pparalx!=NULL)
		*pparalx=paralx;

	return(Vout);
}


double	MoonBody::Age(void)
{
	age=fabs(((Coord[Ephemeris::GEO].Spher[Vector::COORD_TETA] > pE->Sun.Coord[Ephemeris::GEO].Spher[Vector::COORD_TETA]?phase:-phase)-180.0)/360.0*MOO_MONTH)/*+1.0*/;

	return(age);
}

//////////////////////////////////////////////////////////////////////////////
void	Nutation::Compute(Ephemeris *pEphe)
{
	static const double	Vlam[13][5]=NUT_L_VV,Veps[9][5]=NUT_E_VV;
	static const VectData	Vlc1[4]=NUT_L_VC1,Vec1[2]=NUT_E_VC1;
	static const double Vlc2[9]=NUT_L_VC2,Vec2[7]=NUT_E_VC2;
	VectData	V;
	unsigned short int	i;


	pE=pEphe;

	V.d[0]=DEG2RAD(pE->Sun.OrbitElem.L);
	V.d[1]=DEG2RAD(pE->Sun.OrbitElem.M);
	V.d[2]=DEG2RAD(pE->Moon.OrbitElem.L);
	V.d[3]=DEG2RAD(pE->Moon.OrbitElem.M);
	V.d[4]=pE->Moon.OrbitElem.OmegaR;

	// Longitude Nutation
	lamda=0;

	for(i=0;i<4;i++)
		lamda+=Vlc1[i].DotT(pE->Observ.Inst.t)*sin(V.Dot5(Vlam[i]));

	for(i=4;i<13;i++)
		lamda+=Vlc2[i-4]*sin(V.Dot5(Vlam[i]));

	// Obliquity Nutation
	epsilon=Vec1[0].DotT(pE->Observ.Inst.t)*cos(V.Dot5(Veps[0]));
	epsilon+=Vec1[1].DotT(pE->Observ.Inst.t)*cos(V.Dot5(Veps[1]));

	for(i=2;i<9;i++)
		epsilon+=Vec2[i-2]*cos(V.Dot5(Veps[i]));

	// Convert arc second to degree
	lamda/=3600.0;
	epsilon/=3600.0;
}
