/* Jean Philippe EIMER - 2014 *
 *      phil@jpmr.org         */

#include <math.h>
#include <wx/textfile.h>

#include "include.h"


#define	EAR_RT_M	(EAR_RT*1000.0)
#define	MAX_TZ	(12 * 3600L)

#ifndef __WXMSW__
	#define	PARAM_FILENAME	wxT("/.wxepherc")
	#define	PARAM_BACKUP_FILENAME	wxT("/.wxepherc~")
#else	// __WXMSW__
	#define	PARAM_FILENAME	wxT("\\wxEphe\\wxepherc")
	#define	PARAM_BACKUP_FILENAME	wxT("\\wxEphe\\wxepherc.bak")
#endif	// __WXMSW__

#define	XML_ROOT			wxT("wxEpheRC")
#define	XML_LOCATION_NAME	wxT("location_name")
#define	XML_LOCATION_LONG	wxT("location_longitude")
#define	XML_LOCATION_LAT	wxT("location_latitude")
#define	XML_LOCATION_ALT	wxT("location_altitude")
#define	XML_LOCATION_TZ		wxT("location_time_zone")
#define	XML_LOCATION_REFRAC	wxT("refraction")

#define	DEFAULT_REFRACTION	0.61


float	ReadSexa(wxString &line, const wxUniChar sep = wxT(' ') )
{
	double	l0, l;
	float	result;


	line.ToDouble(&l0);
	result = fabs(l0);

	line = line.AfterFirst( sep );
	line.ToDouble(&l);
	result += l/60.0;

	line = line.AfterFirst( sep );
	line.ToDouble(&l);
	result += l/3600.0;

	result = copysign(result, l0);

	return(result);
}


wxString	FormatSexa(const wxString &in)
{
	double	d;
	long	i;
	unsigned char	m, s;


	in.ToDouble(&d);
	i = (long)d;

	d -= (double)i;
	d = fabs(d*100.0);
	m = (unsigned char)d;

	d -= (double)m;
	d = fabs(d*100.0);
	s = (unsigned char)d;

	return( wxString::Format( wxT("%+ld:%02u:%02u"), i, m, s ) );
}


bool	Observer::LoadLocation(void)
{
	wxTextFile	*file;
	wxString filename;
	wxString line, word;
	bool	ret = false;
	double	li;


#ifndef __WXMSW__
	if( !wxGetEnv(wxT("HOME"), &filename ) )
#else	// __WXMSW__
	if( !wxGetEnv(wxT("APPDATA"), &filename ) )
#endif	// __WXMSW__
		filename = wxT(".");

	filename.Append( PARAM_FILENAME );

	file = new wxTextFile(filename);

	if( file->Exists() && file->Open() )
	{
		line = file->GetFirstLine();

		while( !file->Eof() && line.GetChar(0) == wxT('#') )		// filter comments
			line = file->GetNextLine();

		param.location_name = line.BeforeFirst( wxT('\t') );

		line = line.AfterFirst( wxT('\t') );

		word = line.BeforeFirst( wxT('\t') );
		param.location_long = ReadSexa(word);

		line = line.AfterFirst( wxT('\t') );

		word = line.BeforeFirst( wxT('\t') );
		param.location_lat = ReadSexa(word);

		line = line.AfterFirst( wxT('\t') );

		word = line.BeforeFirst( wxT('\t') );
		word.ToDouble(&li);
		param.location_alt = (int)li;

		line = line.AfterFirst( wxT('\t') );

		word = line.BeforeFirst( wxT('\t') );
		word.ToDouble(&li);
		param.location_zone = (long)(li * 3600.0);

		param.refraction = DEFAULT_REFRACTION;
		eta1 = RAD2DEG(acos(EAR_RT_M/(EAR_RT_M+param.location_alt)));
		eta2L = RAD2DEG(0);	// atan(h/d);
		eta2C = RAD2DEG(0);	// atan(h/d);

		ret = true;
	}	// if open

	delete(file);

	return( ret );
}


bool Observer::LoadLocationXML(void)
{
	wxXmlDocument	xml;
	wxXmlNode	*child;
	wxString	filename, s1, s2;
	long	l;
	double	d;
	unsigned char	cpt = 0;
	bool	ret = false;


#ifndef __WXMSW__
	if( !wxGetEnv(wxT("HOME"), &filename ) )
#else	// __WXMSW__
	if( !wxGetEnv(wxT("APPDATA"), &filename ) )
#endif	// __WXMSW__
		filename = wxT(".");

	filename.Append( PARAM_FILENAME );

	if( wxTextFile( filename ).Exists() &&  xml.Load( filename ) && xml.GetRoot()->GetName() == XML_ROOT )
	{
		child = xml.GetRoot()->GetChildren();

		while( child )
		{
			s1 = child->GetName();
			s2 = child->GetNodeContent();
			s2.ToLong(&l, 10);
			s2.ToDouble(&d);

			if ( s1 == XML_LOCATION_NAME )
			{
				param.location_name = s2;
				cpt++;
			}
			else if ( s1 == XML_LOCATION_LONG )
			{
				param.location_long = ReadSexa( s2, wxT(':') );
				cpt++;
			}
			else if ( s1 == XML_LOCATION_LAT )
			{
				param.location_lat = ReadSexa( s2, wxT(':') );
				cpt++;
			}
			else if ( s1 == XML_LOCATION_ALT )
			{
				param.location_alt = (int)l;
				cpt++;
			}
			else if ( s1 == XML_LOCATION_TZ )
			{
				param.location_zone = (long)(d * 3600.0);
				cpt++;
			}

			else if ( s1 == XML_LOCATION_REFRAC )
			{
				param.refraction = (float)d;
				cpt++;
			}

			child = child->GetNext();
		}	// while child

		if( cpt != PARAMETER_NB )
			wxLogMessage( wxString::Format( _("%u missing parameter(s) !"), PARAMETER_NB - cpt) );
		else
		{
			eta1 = RAD2DEG(acos(EAR_RT_M/(EAR_RT_M+param.location_alt)));
			eta2L = RAD2DEG(0);	// atan(h/d);
			eta2C = RAD2DEG(0);	// atan(h/d);
			ret = true;
		}
	}	// if xml.load

	return( ret );
}


bool Observer::SaveLocationXML( const wxString &n, const wxString &lng, const wxString &lat, int alt, long z )
{
	wxXmlDocument	xml;
	wxXmlNode	*root, *child;
	wxString	filename, filename_backup;
	bool	ret = false;


#ifndef __WXMSW__
	if( !wxGetEnv(wxT("HOME"), &filename ) )
#else	// __WXMSW__
	if( !wxGetEnv(wxT("APPDATA"), &filename ) )
#endif	// __WXMSW__
		filename = wxT(".");

	filename_backup = filename;
	filename.Append( PARAM_FILENAME );

	// backup previous version
	if( wxTextFile( filename ).Exists() )
	{
		filename_backup.Append( PARAM_BACKUP_FILENAME );
		wxRenameFile( filename, filename_backup );
	}
#ifdef	__WXMSW__
	else if( !wxFileName( filename ).DirExists() )
		wxFileName::Mkdir( wxFileName( filename ).GetPath(), wxS_DIR_DEFAULT , wxPATH_MKDIR_FULL );		// create data directory if it doesn't exist
#endif	// __WXMSW__

	root = new wxXmlNode( NULL, wxXML_ELEMENT_NODE, XML_ROOT );
	xml.SetRoot(root);

	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_NAME );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), n ) );
	root->AddChild(child);
	root->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxT("\n") ) );

	root->AddChild( new wxXmlNode( wxXML_COMMENT_NODE, wxT(""), wxT(" dd:mm:ss , minus for East ") ) );
	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_LONG );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), FormatSexa( lng ) ) );
	root->AddChild(child);
	root->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxT("\n") ) );

	root->AddChild( new wxXmlNode( wxXML_COMMENT_NODE, wxT(""), wxT(" dd:mm:ss , minus for South ") ) );
	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_LAT );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), FormatSexa( lat ) ) );
	root->AddChild(child);
	root->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxT("\n") ) );

	root->AddChild( new wxXmlNode( wxXML_COMMENT_NODE, wxT(""), wxT(" unit is 'meter' ") ) );
	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_ALT );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxString::Format( wxT("%d"), alt ) ) );
	root->AddChild(child);
	root->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxT("\n") ) );

	root->AddChild( new wxXmlNode( wxXML_COMMENT_NODE, wxT(""), wxT(" unit is 'hour' ") ) );
	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_TZ );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxString::Format( wxT("%+.2f"), (float)z/100.0 ) ) );
	root->AddChild(child);
	root->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxT("\n") ) );

	root->AddChild( new wxXmlNode( wxXML_COMMENT_NODE, wxT(""), wxT(" unit is 'angular degree' ") ) );
	child = new wxXmlNode( wxXML_ELEMENT_NODE, XML_LOCATION_REFRAC );
	child->AddChild( new wxXmlNode( wxXML_TEXT_NODE, wxT(""), wxString::Format( wxT("%.2f"), DEFAULT_REFRACTION ) ) );
	root->AddChild(child);

	xml.Save(filename);

	return( ret );
}


void	Observer::InitLocation(const wxString &n, double lng, double lat, int alt, long z)
{
	param.location_name = wxString(n);
	param.location_long = lng;
	param.location_lat = lat;
	param.location_alt = alt;
	param.location_zone = z;

	param.refraction = DEFAULT_REFRACTION;
	eta1 = RAD2DEG(acos(EAR_RT_M/(EAR_RT_M+param.location_alt)));
	eta2L = RAD2DEG(0);	// atan(h/d);
	eta2C = RAD2DEG(0);	// atan(h/d);
}

/*
double	Observer::InitTime(void)
{
	Inst.tz = wxDateTime::TimeZone( zone );

	Inst.dtlocal.SetToCurrent();
	Inst.dtlocal.SetMillisecond(0);

	Inst.Init(param.location_long);

	return(Inst.tu);
}
*/

double	Observer::InitTime( wxDateTime& wxdt )
{
	Inst.tz = wxDateTime::TimeZone( zone );

	Inst.dtlocal = wxdt;
	Inst.dtlocal.SetMillisecond(0);

	Inst.Init(param.location_long);

	return(Inst.tu);
}


void	Observer::ChangeDST( bool new_dst )
{
	long z = param.location_zone;


	dst = new_dst;

	if( dst )
		z += 3600;

	if( z > MAX_TZ )
		z -= 2*MAX_TZ;

	if( z < -MAX_TZ )
		z += 2*MAX_TZ;

	zone = z;
}

/*
double	Observer::InitTime(unsigned char day, unsigned char month, int year, unsigned char hour, unsigned char min, unsigned char sec, bool btu)
{
	if(btu)		// heure entrée en TU
	{
		Inst.dtlocal.Set(day, (wxDateTime::Month)(month-1), year, hour, min, sec, 0);
		Inst.dtlocal.FromUTC();
	}
	else		// heure entrée comme heure locale
	{
		Inst.dtlocal.Set(day, (wxDateTime::Month)(month-1), year, hour, min, sec, 0);
	}
	Inst.tz = wxDateTime::Local;

	Inst.Init(param.location_long, zone, btu);

	return(Inst.tu);
}
*/

void	Observer::Display(wxEpheFrameOL *Frame) const
{
	Frame->ObsLocation->ChangeValue(param.location_name);
	Frame->LongitudeDir->SetLabel( param.location_long > 0 ? _(L"W"):_(L"E") );
	Frame->Longitude->ChangeValue( wxString::Format( wxT("%+.4f"), Deci2Sexa(param.location_long) ) );
	Frame->LatitudeDir->SetLabel( param.location_lat > 0 ? _(L"N"):_(L"S") );
	Frame->Latitude->ChangeValue( wxString::Format( wxT("%+.4f"), Deci2Sexa(param.location_lat) ) );
	Frame->Altitude->ChangeValue( wxString::Format( wxT("%u"), param.location_alt ) );
	Frame->TimeZoneValue->ChangeValue( wxString::Format( wxT("%+.4ld"), (zone-(dst?3600:0))/36 ) );
}
